import pkg_resources
import pytest
from unittest.mock import Mock

import itertools
import visidata

# test separately as needed

inputLines = { 'save-sheet': 'tests/jetsam.csv',  # save to some tmp file
                 'open-file': 'tests/jetsam.csv',  # reopen what was just saved ('o' must come after ^S in the commands list)
                 'save-col': 'tests/flotsam.csv',
                'pyobj-expr': '2+2',            # open the python object for '4'
                 'search-col': 'foo',
                 'searchr-col': 'bar',
                 'select-col-regex': '.',
                 'select-cols-regex': '.',
                 'unselect-col-regex': '.',
                 'unselect-cols-regex': '.',
#                 'e': '',               # no change should not error
                 'go-col-regex': 'Unit',          # column name in sample
                 'go-col-number': '2',
                 'go-row-number': '5',              # go to row 5
                 'addcol-bulk': '1',
                 'addcol-expr': 'Unit',          # just copy the column
                 'split-col': '-',
                 'show-expr': 'OrderDate',
                 'setcol-expr': 'OrderDate',
                 'setcell-expr': 'OrderDate',
                 'setcol-range': 'range(100)',
                 'capture-col': '(.)(.*)',
                 'addcol-subst': r'Unit/(\w)/\1', # the first character
                 'search-cols': 'foo',
                 'searchr-cols': 'bar',
                 'select-cols-regex': '.',
                 'unselect-cols-regex': '.',
                 'random-rows': '3',
              }

@pytest.mark.usefixtures('curses_setup')
class TestCommands:

    def test_baseCommands(self, mock_screen):
        'exec each global command at least once'

        cmdlist = visidata.vd.commands

        vs = visidata.Sheet('test_commands')
        vs.reload()
        vd = visidata.vd
        nerrs = 0
        ntotal = 0
        for longname in cmdlist.keys():
            if 'Sheet' not in cmdlist[longname]:
                continue
            ntotal += 1
            print(longname)
            self.runOneTest(mock_screen, longname)
            vd.sync()
            if vd.lastErrors:
                # longname, execstr, and vd.lastErrors
                print("{0} FAILED: {1}\n\n\n {2}".format(longname, cmdlist[longname]['Sheet'].execstr, '\n'.join('\n'.join(x) for x in vd.lastErrors)))
                nerrs += 1
                break
            vs.checkCursor()
        print('%s/%s commands had errors' % (nerrs, ntotal))

    def runOneTest(self, mock_screen, longname):
            visidata.vd.clearCaches()  # we want vd to return a new VisiData object for each command
            vd = visidata.vd
            vd.scr = mock_screen

            if longname in inputLines:
                line = [ch for ch in inputLines[longname]] + ['^J']
                vd.getkeystroke = Mock(side_effect=line)
            else:
                vd.getkeystroke = Mock(side_effect=['^J'])

            sample_file = pkg_resources.resource_filename('visidata', '../sample_data/sample.tsv')
            vs = visidata.TsvSheet('test_commands', source=visidata.Path(sample_file))
            vs.reload.__wrapped__(vs)
            vs.vd = vd
            vd.sheets = [vs]
            vs.mouseX, vs.mouseY = (4, 4)
            vs.draw(mock_screen)
            vs.execCommand(longname, vdglobals=vars(visidata))



from copy import copy

from visidata import Sheet, TsvSheet, options, vd, VisiData

@VisiData.api
def open_usv(vd, p):
    vs = TsvSheet(p.name, source=p)
    vs.options.delimiter = '\u241e'
    vs.options.row_delimiter = '\u241f'
    return vs

@VisiData.api
def save_usv(vd, p, vs):
    usvs = copy(vs)
    usvs.rows = vs.rows
    usvs.options.delimiter = '\u241e'
    usvs.options.row_delimiter = '\u241f'
    vd.save_tsv(p, usvs)



import codecs
import tarfile
import zipfile

from visidata import vd, VisiData, asyncthread, Sheet, Progress, Menu, options
from visidata import ColumnAttr, Column, date, datetime, Path

@VisiData.api
def open_zip(vd, p):
    return vd.ZipSheet(p.name, source=p)

@VisiData.api
def open_tar(vd, p):
    return TarSheet(p.name, source=p)

VisiData.open_tgz = VisiData.open_tar
VisiData.open_txz = VisiData.open_tar
VisiData.open_tbz2 = VisiData.open_tar

@VisiData.api
class ZipSheet(Sheet):
    'Wrapper for `zipfile` library.'
    rowtype = 'files' # rowdef ZipInfo
    columns = [
        ColumnAttr('filename'),
        ColumnAttr('file_size', type=int),
        Column('date_time', type=date,
               getter=lambda col, row: datetime.datetime(*row.date_time)),
        ColumnAttr('compress_size', type=int)
    ]

    def openZipFile(self, fp, *args, **kwargs):
        '''Use VisiData input to handle password-protected zip files.'''
        try:
            return fp.open(*args, **kwargs)
        except RuntimeError as err:
            if 'password required' in err.args[0]:
                pwd = vd.input(f'{args[0].filename} is encrypted, enter password: ', display=False)
                return fp.open(*args, **kwargs, pwd=pwd.encode('utf-8'))
            vd.error(err)

    def openRow(self, fi):
            decodedfp = codecs.iterdecode(self.openZipFile(self.zfp, fi),
                                          encoding=options.encoding,
                                          errors=options.encoding_errors)
            return vd.openSource(Path(fi.filename, fp=decodedfp, filesize=fi.file_size), filetype=options.filetype)

    @asyncthread
    def extract(self, *rows, path=None):
        self.zfp.extractall(members=[r.filename for r in rows], path=path)

    @property
    def zfp(self):
        return zipfile.ZipFile(str(self.source), 'r')

    def iterload(self):
        with self.zfp as zf:
            for zi in Progress(zf.infolist()):
                yield zi


class TarSheet(Sheet):
    'Wrapper for `tarfile` library.'
    rowtype = 'files' # rowdef TarInfo
    columns = [
        ColumnAttr('name'),
        ColumnAttr('size', type=int),
        ColumnAttr('mtime', type=date),
        ColumnAttr('type', type=int),
        ColumnAttr('mode', type=int),
        ColumnAttr('uname'),
        ColumnAttr('gname')
    ]

    def openRow(self, fi):
            tfp = tarfile.open(name=str(self.source))
            decodedfp = codecs.iterdecode(tfp.extractfile(fi),
                                          encoding=options.encoding,
                                          errors=options.encoding_errors)
            return vd.openSource(Path(fi.name, fp=decodedfp, filesize=fi.size))

    def iterload(self):
        with tarfile.open(name=str(self.source)) as tf:
            for ti in Progress(tf.getmembers()):
                yield ti


ZipSheet.addCommand('x', 'extract-file', 'extract(cursorRow)', 'extract current file to current directory')
ZipSheet.addCommand('gx', 'extract-selected', 'extract(*onlySelectedRows)', 'extract selected files to current directory')
ZipSheet.addCommand('zx', 'extract-file-to', 'extract(cursorRow, path=inputPath("extract to: "))', 'extract current file to given pathname')
ZipSheet.addCommand('gzx', 'extract-selected-to', 'extract(*onlySelectedRows, path=inputPath("extract %d files to: " % nSelectedRows))', 'extract selected files to given directory')

vd.addMenu(Menu('File', Menu('Extract',
        Menu('current file', 'extract-file'),
        Menu('current file to', 'extract-file-to'),
        Menu('selected files', 'extract-selected'),
        Menu('selected files to', 'extract-selected-to'),
    )))

vd.addGlobals({
    'ZipSheet': ZipSheet,
    'TarSheet': TarSheet
})
